/*! @brief Flag for checking if CisInterface.hpp has already been included.*/
#ifndef CISINTERFACE_HPP_
#define CISINTERFACE_HPP_

#include "YggInterface.hpp"

/*! @brief Aliases to preserve names of original structures. */
#define CisInput YggInput
#define CisOutput YggOutput
#define CisRpc YggRpc
#define CisRpcServer YggRpcServer
#define CisRpcClient YggRpcClient
#define CisAsciiFileOutput YggAsciiFileOutput
#define CisAsciiFileInput YggAsciiFileInput
#define CisAsciiTableOutput YggAsciiTableOutput
#define CisAsciiArrrayOutput YggAsciiArrayOutput
#define CisAsciiTableInput YggAsciiTableInput
#define CisAsciiArrayInput YggAsciiArrayInput
#define CisPlyOutput YggPlyOutput
#define CisPlyInput YggPlyInput
#define CisObjOutput YggObjOutput
#define CisObjInput YggObjInput

#endif /*CISINTERFACE_HPP_*/
